""" 
Graphene is based on an octree data structure. We mesh at
the lowest agglomerated level of the octree (L2) and then 
create stitched meshes at each higher level of the tree 
recursively starting from those L2 meshes. This scheme allows
use to rapidly reconfigure small regions that change while 
proofreading.

The goal of meshing is to render each L2 label into a distinct
mesh. However, in order for the stitching to work, as in a flat 
segmentation, we need to extend one voxel into the neighboring 
chunk. 

Each L2 chunk represents a self contained proofread agglomeration
of the L1 watershed regions over an area of typically 256x256x512
voxels. Actual proofreading extends across L2 chunks, but this is
represented by higher levels of the octree. Therefore, adjacent
L2 chunks have completely different labels.

In order to create accurate merging between chunks, we'll need to
relabel the overlap region using the lowest parent chunk ids between
the primary chunk and each individual overlap region (therefore each
overlap region may require a different chunk graph level). The reason
we can't just label them all with the root ID and mesh is that levels
higher than the lowest level may introduce non-local information. For
instance, a self touch on the boundary that would have been kept apart
at L3 would be unified at the root. The chunk position determines what 
the lowest root is. For example, a chunk located at the origin will 
require an L3 overlap merge. A chunk located one unit from the origin
will require an L4 merge. A chunk located three units from the origin 
will require an L5 merge, and so on according to the fan-out of the octree.

Once the chunk is relabeled using the correct parent, some odd cases 
crop up.  For example, due to information in the overlap region, L2 
labels that were formerly independent connected components may be 
merged via a chain of one or more L3+ relabeled overlap labels. Therefore,
we apply the following rules:

1. If the L2 label does not touch the overlap region, use that label.
2. Each overlap supervoxel is merged with neighboring labels that share 
  a parent. Each resulting connected component is remapped using the minimum 
  covered L2 label from the primary chunk. If no such label exists it 
  is mapped to 0.

"""
from typing import Sequence
import numpy as np
import time
import collections
from functools import lru_cache
import datetime
import pytz

import networkx as nx

import cc3d

from tqdm import tqdm

import cloudvolume
from cloudvolume.lib import Bbox, Vec, xyzrange
from cloudvolume import CloudVolume
import fastremap

UTC = pytz.UTC

def remap_segmentation(
  cv, chunk_x, chunk_y, chunk_z, mip=2, 
  overlap_vx=1, time_stamp=None, progress=False
):
  ws_cv = CloudVolume(
    cv.meta.cloudpath, mip=mip, 
    progress=progress, fill_missing=cv.fill_missing
  )
  mip_diff = mip - cv.meta.watershed_mip

  mip_chunk_size = np.array(
    cv.meta.graph_chunk_size, dtype=np.int
  ) / np.array([2 ** mip_diff, 2 ** mip_diff, 1])
  mip_chunk_size = mip_chunk_size.astype(np.int)

  offset = Vec(chunk_x, chunk_y, chunk_z) * mip_chunk_size
  bbx = Bbox(offset, offset + mip_chunk_size + overlap_vx)
  if cv.meta.chunks_start_at_voxel_offset:
    bbx += ws_cv.voxel_offset
  bbx = Bbox.clamp(bbx, ws_cv.bounds)

  seg = ws_cv[bbx][..., 0]

  if not np.any(seg):
    return seg

  sv_remapping, unsafe_dict = get_lx_overlapping_remappings(
    cv, chunk_x, chunk_y, chunk_z, 
    time_stamp=time_stamp, progress=progress
  )

  seg = fastremap.mask_except(seg, list(sv_remapping.keys()), in_place=True)
  fastremap.remap(seg, sv_remapping, preserve_missing_labels=True, in_place=True)

  for unsafe_root_id in tqdm(unsafe_dict.keys(), desc="Unsafe Relabel", disable=(not progress)):
    bin_seg = seg == unsafe_root_id

    if np.sum(bin_seg) == 0:
        continue

    cc_seg = cc3d.connected_components(bin_seg)
    for i_cc in range(1, np.max(cc_seg) + 1):
      bin_cc_seg = cc_seg == i_cc

      overlaps = []
      overlaps.extend(np.unique(seg[-2, :, :][bin_cc_seg[-1, :, :]]))
      overlaps.extend(np.unique(seg[:, -2, :][bin_cc_seg[:, -1, :]]))
      overlaps.extend(np.unique(seg[:, :, -2][bin_cc_seg[:, :, -1]]))
      overlaps = np.unique(overlaps)

      linked_l2_ids = overlaps[np.in1d(overlaps, unsafe_dict[unsafe_root_id])]

      if len(linked_l2_ids) == 0:
        seg[bin_cc_seg] = 0
      else:
        seg[bin_cc_seg] = linked_l2_ids[0]

  return seg

@lru_cache(maxsize=200)
def get_higher_to_lower_remapping(cv, chunk_id, time_stamp):
  """ 
  Retrieves lx node id to sv id mappping

  :param cv: CloudVolumeGraphene object
  :param chunk_id: np.uint64
  :param time_stamp: datetime object
  :return: dictionary
  """
  rr_chunk = cv.get_chunk_mappings(chunk_id=chunk_id, timestamp=time_stamp)

  # This for-loop ensures that only the latest lx_ids are considered
  # The order by id guarantees the time order (only true for same neurons
  # but that is the case here).
  lx_remapping = {}
  all_lower_ids = set()
  for k in sorted(rr_chunk.keys(), reverse=True):
    this_child_ids = rr_chunk[k]

    if this_child_ids[0] in all_lower_ids:
      continue

    all_lower_ids.update(list(this_child_ids))
    lx_remapping[k] = this_child_ids

  return lx_remapping


@lru_cache(maxsize=None)
def get_root_lx_remapping(cv, chunk_id, stop_layer, time_stamp):
  """
  Retrieves root to l2 node id mapping

  :param cv: CloudVolumeGraphene object
  :param chunk_id: np.uint64
  :param stop_layer: int
  :param time_stamp: datetime object
  :return: multiples
  """
  lx_id_remap = get_higher_to_lower_remapping(
    cv, chunk_id, time_stamp=time_stamp
  )

  lx_ids = np.array(list(lx_id_remap.keys()))
  root_ids = cv.get_roots(lx_ids, stop_layer=stop_layer)

  return lx_ids, np.array(root_ids), lx_id_remap


# @lru_cache(maxsize=None)
def get_lx_overlapping_remappings(
  cv, chunk_x, chunk_y, chunk_z, 
  time_stamp=None, progress=False
):
  """ 
  Retrieves sv id to layer mapping for chunk with overlap in positive
    direction (one chunk)

  :param cv: CloudVolumeGraphene object
  :param chunk_x: np.uint64
  :param chunk_y: np.uint64
  :param chunk_z: np.uint64
  :param time_stamp: datetime object
  :return: multiples
  """
  if time_stamp is None:
    time_stamp = datetime.datetime.utcnow()

  if isinstance(time_stamp, datetime.datetime) and time_stamp.tzinfo is None:
    time_stamp = UTC.localize(time_stamp)

  chunk_layer = 2
  chunk_id = cv.meta.encode_label(chunk_layer, chunk_x, chunk_y, chunk_z, 0)

  neigh_chunk_ids = []
  neigh_parent_chunk_ids = []

  # Collect neighboring chunks and their parent chunk ids
  # We only need to know about the parent chunk ids to figure the lowest
  # common chunk
  # Notice that the first neigh_chunk_id is equal to `chunk_id`.
  for x in range(chunk_x, chunk_x + 2):
    for y in range(chunk_y, chunk_y + 2):
      for z in range(chunk_z, chunk_z + 2):
        
        neigh_chunk_id = cv.meta.encode_label(layer=2, x=x, y=y, z=z, segid=0)
        neigh_chunk_ids.append(neigh_chunk_id)
        neigh_parent_chunk_ids.append(
          get_parent_chunk_ids(cv, neigh_chunk_id)
        )

  # Find lowest common chunk
  # finding the layer that has the lowest common parent of
  # these 8 neighboring chunks
  # Layer agreement will equal 0 if that is layer 3, 1 if
  # that is layer 4, etc
  neigh_parent_chunk_ids = np.array(neigh_parent_chunk_ids)
  layer_agreement = np.all(
    (neigh_parent_chunk_ids - neigh_parent_chunk_ids[0]) == 0, axis=0
  )
  stop_layer = np.where(layer_agreement)[0][0] + chunk_layer

  # Find the parent in the lowest common chunk for each l2 id. These parent
  # ids are referred to as root ids even though they are not necessarily the
  # root id.
  neigh_lx_ids = []
  neigh_lx_id_remap = {}
  neigh_root_ids = []

  safe_lx_ids = []
  unsafe_lx_ids = []
  unsafe_root_ids = []

  # This loop is the main bottleneck
  for neigh_chunk_id in tqdm(neigh_chunk_ids, disable=(not progress), desc="Neighbor"):
    # print(f"Neigh: {neigh_chunk_id} --------------")

    lx_ids, root_ids, lx_id_remap = get_root_lx_remapping(
      cv, neigh_chunk_id, stop_layer, time_stamp=time_stamp
    )
    neigh_lx_ids.extend(lx_ids)
    neigh_lx_id_remap.update(lx_id_remap)
    neigh_root_ids.extend(root_ids)

    if neigh_chunk_id == chunk_id:
      # The first neigh_chunk_id is the one we are interested in. All lx
      # ids that share no root id with any other lx id are "safe", meaning
      # that we can easily obtain the complete remapping (including
      # overlap) for these. All other ones have to be resolved using the
      # segmentation.
      _, u_idx, c_root_ids = np.unique(
        neigh_root_ids, return_counts=True, return_index=True
      )

      safe_lx_ids = lx_ids[u_idx[c_root_ids == 1]]
      unsafe_lx_ids = lx_ids[~np.in1d(lx_ids, safe_lx_ids)]
      unsafe_root_ids = np.unique(root_ids[u_idx[c_root_ids != 1]])

  lx_root_dict = dict(zip(neigh_lx_ids, neigh_root_ids))
  root_lx_dict = collections.defaultdict(list)

  # Future sv id -> lx mapping
  sv_ids = []
  lx_ids_flat = []

  # Do safe ones first
  for i_root_id in range(len(neigh_root_ids)):
    root_lx_dict[neigh_root_ids[i_root_id]].append(neigh_lx_ids[i_root_id])

  for lx_id in safe_lx_ids:
    root_id = lx_root_dict[lx_id]
    for neigh_lx_id in root_lx_dict[root_id]:
      lx_sv_ids = neigh_lx_id_remap[neigh_lx_id]
      sv_ids.extend(lx_sv_ids)
      lx_ids_flat.extend([lx_id] * len(neigh_lx_id_remap[neigh_lx_id]))

  # For the unsafe ones we can only do the in chunk svs
  # But we will map the out of chunk svs to the root id and store the
  # hierarchical information in a dictionary
  for lx_id in unsafe_lx_ids:
    sv_ids.extend(neigh_lx_id_remap[lx_id])
    lx_ids_flat.extend([lx_id] * len(neigh_lx_id_remap[lx_id]))

  unsafe_dict = collections.defaultdict(list)
  for root_id in unsafe_root_ids:
    if np.sum(~np.in1d(root_lx_dict[root_id], unsafe_lx_ids)) == 0:
      continue

    for neigh_lx_id in root_lx_dict[root_id]:
      unsafe_dict[root_id].append(neigh_lx_id)

      if neigh_lx_id in unsafe_lx_ids:
        continue

      sv_ids.extend(neigh_lx_id_remap[neigh_lx_id])
      lx_ids_flat.extend([root_id] * len(neigh_lx_id_remap[neigh_lx_id]))

  # Combine the lists for a (chunk-) global remapping
  sv_remapping = dict(zip(sv_ids, lx_ids_flat))

  return sv_remapping, unsafe_dict

def get_parent_chunk_ids(cv, label):
  """ 
  Creates list of chunk parent ids

  :param label: np.uint64
  :return: np.ndarray
  """
  parent_chunk_layers = range(
    int(cv.meta.decode_layer_id(label)) + 1, int(cv.meta.n_layers) + 1
  )

  chunk_coord = Vec(*cv.meta.decode_chunk_position(label))

  parent_chunk_ids = [ cv.meta.decode_chunk_id(label) ]
  for layer in parent_chunk_layers:
    chunk_coord = chunk_coord // cv.meta.fan_out
    chunk_id = cv.meta.encode_label(
      layer, chunk_coord[0], chunk_coord[1], chunk_coord[2], segid=0
    )
    parent_chunk_ids.append(chunk_id)

  return np.array(parent_chunk_ids, dtype=np.uint64)



# @license
# Copyright 2016,2017 The Neuroglancer Authors
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math

import numpy as np
from cloudvolume import CloudVolume, Vec
from cloudvolume.lib import min2, getprecision

DEFAULT_MAX_DOWNSAMPLING = 64 # maximum factor to downsample by
DEFAULT_MAX_DOWNSAMPLED_SIZE = 128 # minimum length of a side after downsampling
DEFAULT_MAX_DOWNSAMPLING_SCALES = float('inf')

def scale_series_to_downsample_factors(scales):
  fullscales = [ np.array(scale) for scale in scales ] 
  factors = []
  for i in range(1, len(fullscales)):
    factors.append( fullscales[i] / fullscales[i - 1]  )
  return [ factor.astype(int) for factor in factors ]


def compute_near_isotropic_downsampling_scales(size,
                                               voxel_size,
                                               dimensions_to_downsample,
                                               max_scales=DEFAULT_MAX_DOWNSAMPLING_SCALES,
                                               max_downsampling=DEFAULT_MAX_DOWNSAMPLING,
                                               max_downsampled_size=DEFAULT_MAX_DOWNSAMPLED_SIZE):
    """Compute a list of successive downsampling factors."""

    num_dims = len(voxel_size)
    cur_scale = np.ones((num_dims, ), dtype=int)
    scales = [tuple(cur_scale)]
    while (len(scales) < max_scales 
            and (np.prod(cur_scale) < max_downsampling) 
            and np.all(size / cur_scale > max_downsampled_size)
            and np.all(np.mod(size,cur_scale) == 0)): # the number of chunks should be integer
        # Find dimension with smallest voxelsize.
        cur_voxel_size = cur_scale * voxel_size
        smallest_cur_voxel_size_dim = dimensions_to_downsample[np.argmin(cur_voxel_size[
            dimensions_to_downsample])]
        cur_scale[smallest_cur_voxel_size_dim] *= 2
        target_voxel_size = cur_voxel_size[smallest_cur_voxel_size_dim] * 2
        for d in dimensions_to_downsample:
            if d == smallest_cur_voxel_size_dim:
                continue
            d_voxel_size = cur_voxel_size[d]
            if abs(d_voxel_size - target_voxel_size) > abs(d_voxel_size * 2 - target_voxel_size):
                cur_scale[d] *= 2
        scales.append(tuple(cur_scale))
    return scales

def compute_two_dimensional_near_isotropic_downsampling_scales(
        size,
        voxel_size,
        max_scales=float('inf'),
        max_downsampling=DEFAULT_MAX_DOWNSAMPLING,
        max_downsampled_size=DEFAULT_MAX_DOWNSAMPLED_SIZE):
    """Compute a list of successive downsampling factors for 2-d tiles."""

    max_scales = min(max_scales, 10)

    # First compute a set of 2-d downsamplings for XY, XZ, and YZ with a high
    # number of max_scales, and ignoring other criteria.
    scales_transpose = [
        compute_near_isotropic_downsampling_scales(
            size=size,
            voxel_size=voxel_size,
            dimensions_to_downsample=dimensions_to_downsample,
            max_scales=max_scales,
            max_downsampling=float('inf'),
            max_downsampled_size=0, ) for dimensions_to_downsample in [[0, 1], [0, 2], [1, 2]]
    ]

    # Truncate all list of scales to the same length, once the stopping criteria
    # is reached for all values of dimensions_to_downsample.
    scales = [((1, ) * 3, ) * 3]
    size = np.array(size)

    def scale_satisfies_criteria(scale):
        return np.prod(scale) < max_downsampling and (size / scale).max() > max_downsampled_size

    for i in range(1, max_scales):
        cur_scales = tuple(scales_transpose[d][i] for d in range(3))
        if all(not scale_satisfies_criteria(scale) for scale in cur_scales):
            break
        scales.append(cur_scales)
    return scales

def compute_plane_downsampling_scales(size, preserve_axis='z',
                                       max_scales=DEFAULT_MAX_DOWNSAMPLING_SCALES,
                                       max_downsampling=DEFAULT_MAX_DOWNSAMPLING,
                                       max_downsampled_size=DEFAULT_MAX_DOWNSAMPLED_SIZE):

    axis_map = { 'x': 0, 'y': 1, 'z': 2 }
    preserve_axis = axis_map[preserve_axis]

    size = np.array(size)

    if np.any(size <= 0):
        return [ (1,1,1) ]

    size[preserve_axis] = size[ (preserve_axis + 1) % 3 ]

    dimension = min(*size)
    num_downsamples = int(np.log2(dimension / max_downsampled_size))
    num_downsamples = min(num_downsamples, max_scales)

    factor = 2
    scales = [ (1,1,1) ]
    for i in range(num_downsamples):
        if factor > max_downsampling:
            break
        elif dimension / factor < max_downsampled_size:
            break

        scale = [ factor, factor, factor ]
        scale[preserve_axis] = 1

        scales.append(tuple(scale))
        factor *= 2

    return scales
    
def compute_factors(ds_shape, factor, chunk_size):
  grid_size = Vec(*ds_shape, dtype=np.float32) / Vec(*chunk_size, dtype=np.float32)
  # find the dimension which will tolerate the smallest number of downsamples and 
  # return the number it will accept. + 0.0001 then truncate to compensate for FP errors
  # that would result in the log e.g. resulting in 1.9999999382 when it should be an
  # exact result.

  # This filtering is to avoid problems with dividing by log(1)
  grid_size = [ g for f, g in zip(factor, grid_size) if f != 1 ]
  grid_size = Vec(*grid_size, dtype=np.float32)

  factor_div = [ f for f in factor if f != 1 ]
  factor_div = Vec(*factor_div, dtype=np.float32)

  if len(factor_div) == 0:
    return []

  N = np.log(grid_size) / np.log(factor_div)
  N += 0.0001
  return [ factor ] * int(min(N))

def axis_to_factor(axis):
  if axis == 'x':
    return (1,2,2)
  elif axis == 'y':
    return (2,1,2)
  elif axis == 'z':
    return (2,2,1)
  else:
    raise ValueError("Axis not supported: " + str(axis))

def compute_scales(vol, mip, shape, axis, factor, chunk_size=None):
  shape = min2(vol.meta.volume_size(mip), shape)
  # sometimes we downsample a base layer of 512x512 
  # into underlying chunks of 64x64 which permits more scales
  underlying_mip = (mip + 1) if (mip + 1) in vol.available_mips else mip

  if chunk_size:
    scale_chunk_size = Vec(*chunk_size).astype(np.float32)
  else:
    scale_chunk_size = vol.meta.chunk_size(underlying_mip).astype(np.float32)

  if factor is None:
    factor = axis_to_factor(axis)

  factors = compute_factors(shape, factor, scale_chunk_size)
  scales = [ vol.meta.resolution(mip) ]

  precision = max(map(getprecision, vol.meta.resolution(mip)))

  def prec(x):
    if precision == 0:
      return int(x)
    return round(x, precision)

  for factor3 in factors:
    scales.append(
      list(map(prec, Vec(*scales[-1], dtype=np.float32) * Vec(*factor3)))
    )
  return scales[1:]

def create_downsample_scales(
    layer_path, mip, ds_shape, axis='z', 
    preserve_chunk_size=False, chunk_size=None,
    encoding=None, factor=None
  ):
  vol = CloudVolume(layer_path, mip)

  scales = compute_scales(vol, mip, ds_shape, axis, factor, chunk_size)

  if len(scales) == 0:
    print("WARNING: No scales generated.")

  for scale in scales:
    vol.meta.add_resolution(scale, encoding=encoding, chunk_size=chunk_size)

  if chunk_size is None:
    if preserve_chunk_size or len(scales) == 0:
      chunk_size = vol.scales[mip]['chunk_sizes']
    else:
      chunk_size = vol.scales[mip + 1]['chunk_sizes']
  else:
    chunk_size = [ chunk_size ]

  if encoding is None:
    encoding = vol.scales[mip]['encoding']

  for i in range(mip + 1, mip + len(scales) + 1):
    vol.scales[i]['chunk_sizes'] = chunk_size

  vol.commit_info()
  return vol


def downsample_shape_from_memory_target(
  data_width, cx, cy, cz, 
  factor, byte_target,
  max_mips=float('inf')
):
  """
  Compute the shape that will give the most downsamples for a given 
  memory target (e.g. 3e9 bytes aka 3 GB).

  data_width: byte size of dtype
  cx: chunk size x
  cy: chunk size y 
  cz: chunk size z
  factor: (2,2,1) or (2,2,2) are supported
  byte_target: memory used should be less than this

  Returns: Vec3 shape
  """
  # formulas come from solving the following optimization equations:
  #
  # factor (1,1,1)
  # find integers n and m such that
  # |n * cx - m * cy| is (approximately) minimized
  # treat cz as fixed to make thing easier.
  # We start with a guess that n = sqrt(byte_target / data_width / cx / cy / cz)
  #
  # factor (2,2,1)
  # 4/3 * data_width * cx^(2^n) * cy^(2^n) * cz < byte_target
  #
  # factor (2,2,2)
  # 8/7 * data_width * cx^(2^n) * cy^(2^n) * cz^(2^n) < byte_target
  #
  # it's possible to solve for an arbitrary factor, but more complicated
  # and we really only need those two as the blowup gets intense.
  if byte_target <= 0:
    raise ValueError(f"Unable to pick a shape for a byte budget <= 0. Got: {byte_target}")

  if cx * cy * cz <= 0:
    raise ValueError(f"Chunk size must have a positive integer volume. Got: <{cx},{cy},{cz}>")

  def n_shape(n, c_):
    num_downsamples = int(math.log2((c_ ** (2*n)) / c_))
    num_downsamples = int(min(num_downsamples, max_mips))
    return c_ * (2 ** num_downsamples)

  if factor == (1,1,1):
    n = int(math.sqrt(byte_target / data_width / cx / cy / cz))
    m = int(n * cx / cy)
    out = Vec(n * cx, m * cy, cz)
  elif factor == (2,2,1):
    if cx * cy == 1:
      size = 2 ** int(math.log2(math.sqrt(byte_target / cz)))
      out = Vec(size, size, cz)
    else:
      n = math.log(3/4 * byte_target / data_width / cz)
      n = n / 2 / math.log(cx * cy)
      shape = lambda c_: n_shape(n, c_) 
      out = Vec(shape(cx), shape(cy), cz)
  elif factor == (2,2,2):
    if cx * cy * cz == 1:
      size = 2 ** int(math.log2(round(byte_target ** (1/3), 5)))
      out = Vec(size, size, size)
    else:
      n = math.log(7/8 * byte_target / data_width)
      n = n / 2 / math.log(cx * cy * cz) 
      shape = lambda c_: n_shape(n, c_) 
      out = Vec(shape(cx), shape(cy), shape(cz))
  else:
    raise ValueError(f"This is now a harder optimization problem. Got: {factor}")

  out = out.astype(int)
  min_shape = Vec(cx,cy,cz)
  if any(out < min_shape):
    raise ValueError(
      f"Too little memory allocated to create a valid task."
      f" Got: {byte_target} Predicted Shape: {out} Minimum Shape: {min_shape}"
    )

  return out




import multiprocessing as mp
import random
import signal
import sys
import time
import traceback

import click
from taskqueue import TaskQueue

from igneous import EmptyVolumeException
# from igneous import logger

from igneous.secrets import SQS_URL, SQS_REGION_NAME, SQS_ENDPOINT_URL, LEASE_SECONDS

@click.command()
@click.option('-m', default=False,  help='Run in parallel. DEPRECATED: Please use -p.', is_flag=True)
@click.option('-p', default=1,  help='Run with this number of parallel processes.')
@click.option('--queue', default=SQS_URL,  help='TaskQueue protocol url for queue. e.g. sqs://test-queue or fq:///tmp/test-queue')
@click.option('--region_name', default=SQS_REGION_NAME,  help='AWS region in which the taskqueue resides')
@click.option('--endpoint_url', default=SQS_ENDPOINT_URL,  help='Endpoint of the SQS service if not AWS (NOT the queue url)')
@click.option('--seconds', default=LEASE_SECONDS, help="Lease seconds.")
@click.option('--tally/--no-tally', default=True, help="Tally completed fq tasks.")
@click.option('--loop', default=-1, help='Execute for at least this many seconds. 0: Run only a single task. -1: Loop forever (default).')
def command(m, p, queue, region_name, endpoint_url, seconds, tally, loop):
  loop = float(loop)
  args = (queue, region_name, endpoint_url, seconds, tally, loop)

  if m:
    print("-m is deprecated. please use -p 0 instead.")
    p = mp.cpu_count()

  if p == 1:
    execute(*args)
    return 

  # if multiprocessing
  proc = p if p > 0 else mp.cpu_count()
  pool = mp.Pool(processes=proc)
  print("Running %s threads of execution." % proc)
  try:
    for _ in range(proc):
      pool.apply_async(execute, args)
    pool.close()
    pool.join()
  except KeyboardInterrupt:
    print("Interrupted. Exiting.")
    pool.terminate()
    pool.join()

def execute(queue, region_name, endpoint_url, seconds, tally, loop):
  tq = TaskQueue(queue, region_name=region_name, endpoint_url=endpoint_url)

  print("Pulling from {}".format(tq.qualified_path))
  seconds = int(seconds)

  def stop_after_elapsed_time(elapsed_time):
    if loop < 0:
      return False
    return loop < elapsed_time

  if loop != 0:
    tq.poll(
      lease_seconds=seconds,
      verbose=True,
      tally=tally,
      stop_fn=stop_after_elapsed_time,
    )
  else:
    task = tq.lease(seconds=seconds)
    task.execute()

if __name__ == '__main__':
  command()



import math
import multiprocessing as mp
import os
import sys

import click
from cloudvolume import CloudVolume
import numpy as np
from taskqueue import TaskQueue
from taskqueue.lib import toabs
from taskqueue.paths import get_protocol

from igneous import task_creation as tc
from igneous import downsample_scales
from igneous.secrets import LEASE_SECONDS, SQS_REGION_NAME

from igneous_cli.humanbytes import format_bytes

def normalize_path(queuepath):
  if not get_protocol(queuepath):
    return "fq://" + toabs(queuepath)
  return queuepath

class Tuple3(click.ParamType):
  """A command line option type consisting of 3 comma-separated integers."""
  name = 'tuple3'
  def convert(self, value, param, ctx):
    if isinstance(value, str):
      try:
        value = tuple(map(int, value.split(',')))
      except ValueError:
        self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
      if len(value) != 3:
        self.fail(f"'{value}' does not contain a comma delimited list of 3 integers.")
    return value
  

@click.group()
@click.option("-p", "--parallel", default=1, help="Run with this number of parallel processes. If 0, use number of cores.")
@click.version_option(version="0.5.0")
@click.pass_context
def main(ctx, parallel):
  """
  CLI tool for managing igneous jobs.
  https://github.com/seung-lab/igneous

  Igneous is a tool for producing neuroglancer
  datasets. It scales to hundreds of teravoxels
  or more.

  Select an operation, dataset, and queue and
  tasks will be inserted into the queue. Queues
  can be either SQS or a filesystem directory.

  Then use "igneous execute $queue" to start
  processing that operation.

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version. Run "igneous license" for details.
  """
  parallel = int(parallel)
  if parallel == 0:
    parallel = mp.cpu_count()
  ctx.ensure_object(dict)
  ctx.obj["parallel"] = max(min(parallel, mp.cpu_count()), 1)

@main.command()
def license():
  """Prints the license for this library and cli tool."""
  path = os.path.join(os.path.dirname(__file__), 'LICENSE')
  with open(path, 'rt') as f:
    print(f.read())

@main.command()
@click.argument("path")
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--mip', default=0, help="Build upward from this level of the image pyramid. Default: 0")
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@click.option('--num-mips', default=5, help="Build this many additional pyramid levels. Each increment increases memory requirements per task 4-8x.  Default: 5")
@click.option('--cseg', is_flag=True, default=False, help="Use the compressed_segmentation image chunk encoding scheme. Segmentation only.")
@click.option('--compresso', is_flag=True, default=False, help="Use the compresso image chunk encoding scheme. Segmentation only.")
@click.option('--sparse', is_flag=True, default=False, help="Don't count black pixels in mode or average calculations. For images, eliminates edge ghosting in 2x2x2 downsample. For segmentation, prevents small objects from disappearing at high mip levels.")
@click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of new layers. e.g. 128,128,64")
@click.option('--compress', default=None, help="Set the image compression scheme. Options: 'gzip', 'br'")
@click.option('--volumetric', is_flag=True, default=False, help="Use 2x2x2 downsampling.")
@click.option('--delete-bg', is_flag=True, default=False, help="Issue a delete instead of uploading a background tile. This is helpful on systems that don't like tiny files.")
@click.option('--bg-color', default=0, help="Determines which color is regarded as background. Default: 0")
@click.pass_context
def downsample(
	ctx, path, queue, mip, fill_missing, 
	num_mips, cseg, compresso, sparse, 
	chunk_size, compress, volumetric,
	delete_bg, bg_color
):
  """
  Create an image pyramid for grayscale or labeled images.
  By default, we use 2x2x1 downsampling. 

  The levels of the pyramid are called "mips" (from the fake latin 
  "Multum in Parvo" or "many in small"). The base of the pyramid,
  the highest resolution layer, is mip 0. Each level of the pyramid
  is one mip level higher.

  The general strategy is to downsample starting from mip 0. This
  builds several levels. Once that job is complete, pass in the 
  current top mip level of the pyramid. This builds it even taller
  (referred to as "superdownsampling").
  """
  if cseg and compresso:
    print("igneous: must choose one of --cseg or --compresso")
    sys.exit()

  encoding = None
  if cseg:
    encoding = "compressed_segmentation"
  elif compresso:
    encoding = "compresso"

  factor = (2,2,1)
  if volumetric:
  	factor = (2,2,2)

  tasks = tc.create_downsampling_tasks(
    path, mip=mip, fill_missing=fill_missing, 
    num_mips=num_mips, sparse=sparse, 
    chunk_size=chunk_size, encoding=encoding, 
    delete_black_uploads=delete_bg, 
    background_color=bg_color, 
    compress=compress,
    factor=factor  	
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@main.command()
@click.argument("src")
@click.argument("dest")
@click.option('--queue', default=None, required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue")
@click.option('--mip', default=0, help="Build upward from this level of the image pyramid.", show_default=True)
@click.option('--translate', type=Tuple3(), default=(0, 0, 0), help="Translate the bounding box by X,Y,Z voxels in the new location.")
@click.option('--downsample/--skip-downsample', is_flag=True, default=True, help="Whether or not to produce downsamples from transfer tiles.", show_default=True)
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@click.option('--memory', default=3.5e9, type=int, help="Task memory limit in bytes. Task shape will be chosen to fit and maximize downsamples.", show_default=True)
@click.option('--max-mips', default=5, help="Maximum number of additional pyramid levels.", show_default=True)
@click.option('--cseg', is_flag=True, default=False, help="Use the compressed_segmentation image chunk encoding scheme. Segmentation only.")
@click.option('--compresso', is_flag=True, default=False, help="Use the compresso image chunk encoding scheme. Segmentation only.")
@click.option('--sparse', is_flag=True, default=False, help="Don't count black pixels in mode or average calculations. For images, eliminates edge ghosting in 2x2x2 downsample. For segmentation, prevents small objects from disappearing at high mip levels.")
@click.option('--shape', type=Tuple3(), default=(2048, 2048, 64), help="(overrides --memory) Set the task shape in voxels. This also determines how many downsamples you get. e.g. 2048,2048,64")
@click.option('--chunk-size', type=Tuple3(), default=None, help="Chunk size of destination layer. e.g. 128,128,64")
@click.option('--compress', default="gzip", help="Set the image compression scheme. Options: 'gzip', 'br'", show_default=True)
@click.option('--volumetric', is_flag=True, default=False, help="Use 2x2x2 downsampling.")
@click.option('--delete-bg', is_flag=True, default=False, help="Issue a delete instead of uploading a background tile. This is helpful on systems that don't like tiny files.")
@click.option('--bg-color', default=0, help="Determines which color is regarded as background.", show_default=True)
@click.option('--sharded', is_flag=True, default=False, help="Generate a sharded dataset which reduces the number of files. Downsamples are not generated.")
@click.option('--dest-voxel-offset', type=Tuple3(), default=None, help="Set the voxel offset for this mip level.")
@click.pass_context
def xfer(
	ctx, src, dest, queue, translate, 
  downsample, mip, fill_missing, 
  memory, max_mips, shape, sparse, 
  cseg, compresso, chunk_size, compress, 
  volumetric, delete_bg, bg_color, sharded,
  dest_voxel_offset
):
  """
  Transfer an image layer to another location.

  It is crucial to choose a good task shape. The task
  shape must be a multiple of two of the destination
  image layer chunk size. Too small, and you'll have
  an inefficient transfer. Too big, and you'll run out
  of memory and also have an inefficient transfer.

  Downsamples will by default be automatically calculated
  from whatever material is available. For the default
  2x2x1 downsampling, larger XY dimension is desirable
  compared to Z as more downsamples can be computed for
  each 2x2 increase in the task size.

  Use the --memory flag to automatically compute the
  a reasonable task shape based on your memory limits.
  """
  if cseg and compresso:
    print("igneous: must choose one of --cseg or --compresso")
    sys.exit()

  encoding = None
  if cseg:
    encoding = "compressed_segmentation"
  elif compresso:
    encoding = "compresso"

  factor = (2,2,1)
  if volumetric:
  	factor = (2,2,2)

  if compress and compress.lower() == "false":
    compress = False

  if sharded:
    tasks = tc.create_image_shard_transfer_tasks(
      src, dest,
      chunk_size=chunk_size, fill_missing=fill_missing, mip=mip, 
      dest_voxel_offset=dest_voxel_offset, translate=translate, 
      encoding=encoding, memory_target=memory,
    )
  else:
    tasks = tc.create_transfer_tasks(
      src, dest, 
      chunk_size=chunk_size, fill_missing=fill_missing, 
      dest_voxel_offset=dest_voxel_offset, translate=translate, 
      mip=mip, shape=shape, encoding=encoding, skip_downsamples=(not downsample),
      delete_black_uploads=delete_bg, background_color=bg_color,
      compress=compress, factor=factor, sparse=sparse,
      memory_target=memory, max_mips=max_mips
    )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@main.command()
@click.argument("queue", type=str)
@click.option('--aws-region', default=SQS_REGION_NAME, help=f"AWS region in which the SQS queue resides. Default: {SQS_REGION_NAME}")
@click.option('--lease-sec', default=LEASE_SECONDS, help=f"Seconds to lease a task for. Default: {LEASE_SECONDS}", type=int)
@click.option('--tally/--no-tally', is_flag=True, default=True, help="Tally completed fq tasks. Does not apply to SQS.")
@click.option('--min-sec', default=-1, help='Execute for at least this many seconds and quit after the last task finishes. Special values: (0) Run at most a single task. (-1) Loop forever (default).', type=float)
@click.pass_context
def execute(ctx, queue, aws_region, lease_sec, tally, min_sec):
  """Execute igneous tasks from a queue.

  The queue must be an AWS SQS queue or a FileQueue directory. 
  Examples: (SQS) sqs://my-queue (FileQueue) fq://./my-queue
  (the fq:// is optional).

  See https://github.com/seung-lab/python-task-queue
  """
  parallel = int(ctx.obj.get("parallel", 1))
  args = (queue, aws_region, lease_sec, tally, min_sec)

  if parallel == 1:
    execute_helper(*args)
    return

  pool = mp.Pool(processes=parallel)
  try:
    for _ in range(parallel):
      pool.apply_async(execute_helper, args)
    pool.close()
    pool.join()
  except KeyboardInterrupt:
    print("Interrupted. Exiting.")
    pool.terminate()
    pool.join()

def execute_helper(queue, aws_region, lease_sec, tally, min_sec):
  tq = TaskQueue(normalize_path(queue), region_name=aws_region)

  def stop_after_elapsed_time(elapsed_time):
    if min_sec < 0:
      return False
    return min_sec < elapsed_time

  if min_sec != 0:
    tq.poll(
      lease_seconds=lease_sec,
      verbose=True,
      tally=tally,
      stop_fn=stop_after_elapsed_time,
    )
  else:
    task = tq.lease(seconds=lease_sec)
    task.execute()

@main.group("mesh")
def meshgroup():
  """
  Create 3D meshes from a segmentation. (subgroup)

  Meshing is a two step process of forging then 
  merging. First the meshes are created from a
  regular grid of segmentation cutouts. Second,
  the pieces are glued together.
  """
  pass

@meshgroup.command("forge")
@click.argument("path")
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--mip', default=0, help="Perform meshing using this level of the image pyramid. Default: 0")
@click.option('--shape', type=Tuple3(), default=(448, 448, 448), help="Set the task shape in voxels. Default: 448,448,448")
@click.option('--simplify/--skip-simplify', is_flag=True, default=True, help="Enable mesh simplification. Default: True")
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@click.option('--max-error', default=40, help="Maximum simplification error in physical units. Default: 40 nm")
@click.option('--dust-threshold', default=None, help="Skip meshing objects smaller than this number of voxels within a cutout. No default limit. Typical value: 1000.", type=int)
@click.option('--dir', default=None, help="Write meshes into this directory instead of the one indicated in the info file.")
@click.option('--compress', default=None, help="Set the image compression scheme. Options: 'gzip', 'br'")
@click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.")
@click.pass_context
def mesh_forge(
  ctx, path, queue, mip, shape, 
  simplify, fill_missing, max_error, 
  dust_threshold, dir, compress, 
  spatial_index
):
  """
  (1) Synthesize meshes from segmentation cutouts.

  A large labeled image is divided into a regular
  grid. zmesh is applied to grid point, which performs
  marching cubes and a quadratic mesh simplifier.

  Note that using task shapes with axes less than
  or equal to 511x1023x511 (don't ask) will be more
  memory efficient as it can use a 32-bit mesher.

  zmesh is used: https://github.com/seung-lab/zmesh

  Sharded format not currently supports. Coming soon.
  """
  tasks = tc.create_meshing_tasks(
    path, mip, shape, 
    simplification=simplify, max_simplification_error=max_error,
    mesh_dir=dir, cdn_cache=False, dust_threshold=dust_threshold,
    object_ids=None, progress=False, fill_missing=fill_missing,
    encoding='precomputed', spatial_index=spatial_index, 
    sharded=False, compress=compress
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@meshgroup.command("merge")
@click.argument("path")
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)")
@click.option('--dir', default=None, help="Write manifests into this directory instead of the one indicated in the info file.")
@click.pass_context
def mesh_merge(ctx, path, queue, magnitude, dir):
  """
  (2) Merge the mesh pieces produced from the forging step.

  The per-cutout mesh fragments are then assembled and
  merged. However, this process occurs by compiling 
  a list of fragment files and uploading a "mesh manifest"
  file that is an index for locating the fragments.
  """
  tasks = tc.create_mesh_manifest_tasks(
    path, magnitude=magnitude, mesh_dir=dir
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)


@main.group("skeleton")
def skeletongroup():
  """
  Create skeletons from a segmentation.

  Skeletonizing is a two step process of forging 
  then merging. First the skeletons are created
  from a regular grid of segmentation cutouts.
  Second, the pieces are postprocessed and glued
  together.
  """
  pass

@skeletongroup.command("forge")
@click.argument("path")
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--mip', default=0, help="Perform skeletonizing using this level of the image pyramid. Default: 0")
@click.option('--shape', type=Tuple3(), default=(512, 512, 512), help="Set the task shape in voxels. Default: 512,512,512")
@click.option('--fill-missing', is_flag=True, default=False, help="Interpret missing image files as background instead of failing.")
@click.option('--fix-branching', is_flag=True, default=True, help="Trades speed for quality of branching at forks. Default: True")
@click.option('--fix-borders', is_flag=True, default=True, help="Allows trivial merging of single voxel overlap tasks. Only switch off for datasets that fit in a single task. Default: True")
@click.option('--fix-avocados', is_flag=True, default=False, help="Fixes somata where nuclei and cytoplasm have separate segmentations. Default: False")
@click.option('--fill-holes', is_flag=True, default=False, help="Preprocess each cutout to eliminate background holes and holes caused by entirely contained inclusions. Warning: May remove labels that are considered inclusions. Default: False")
@click.option('--dust-threshold', default=1000, help="Skip skeletonizing objects smaller than this number of voxels within a cutout. Default: 1000.", type=int)
@click.option('--spatial-index/--skip-spatial-index', is_flag=True, default=True, help="Create the spatial index.")
@click.option('--scale', default=4, help="Multiplies invalidation radius by distance from boundary.", type=float)
@click.option('--const', default=10, help="Adds constant amount to invalidation radius in physical units.", type=float)
@click.option('--soma-detect', default=1100, help="Consider objects with peak distances to boundary larger than this soma candidates. Physical units. Default: 1100 nm", type=float)
@click.option('--soma-accept', default=3500, help="Accept soma candidates over this threshold and perform a one-time spherical invalidation around their peak value. Physical units. Default: 3500 nm", type=float)
@click.option('--soma-scale', default=1.0, help="Scale factor for soma invalidation.", type=float)
@click.option('--soma-const', default=300, help="Const factor for soma invalidation.", type=float)
@click.option('--max-paths', default=None, help="Abort skeletonizing an object after this many paths have been traced.", type=float)
@click.option('--sharded', is_flag=True, default=False, help="Generate shard fragments instead of outputing skeleton fragments.")
@click.pass_context
def skeleton_forge(
  ctx, path, queue, mip, shape, 
  fill_missing, dust_threshold, spatial_index,
  fix_branching, fix_borders, fix_avocados, 
  fill_holes, scale, const, soma_detect, soma_accept,
  soma_scale, soma_const, max_paths, sharded
):
  """
  (1) Synthesize skeletons from segmentation cutouts.

  A large labeled image is divided into a regular
  grid. Kimimaro is applied to grid point, which performs
  a TEASAR based skeletonization.

  You can read more about the parameters here:
  https://github.com/seung-lab/kimimaro

  Tutorials are located here:

  - https://github.com/seung-lab/kimimaro/wiki/A-Pictorial-Guide-to-TEASAR-Skeletonization

  - https://github.com/seung-lab/kimimaro/wiki/Intuition-for-Setting-Parameters-const-and-scale

  A guide to how much this might cost is located here:

  - https://github.com/seung-lab/kimimaro/wiki/The-Economics:-Skeletons-for-the-People
  """
  teasar_params = {
    'scale': scale,
    'const': const, # physical units
    'pdrf_exponent': 4,
    'pdrf_scale': 100000,
    'soma_detection_threshold': soma_detect, # physical units
    'soma_acceptance_threshold': soma_accept, # physical units
    'soma_invalidation_scale': soma_scale,
    'soma_invalidation_const': soma_const, # physical units
    'max_paths': max_paths, # default None
  }

  tasks = tc.create_skeletonizing_tasks(
    path, mip, shape, 
    teasar_params=teasar_params, 
    fix_branching=fix_branching, fix_borders=fix_borders, 
    fix_avocados=fix_avocados, fill_holes=fill_holes,
    dust_threshold=dust_threshold, progress=False,
    parallel=1, fill_missing=fill_missing, 
    sharded=sharded, spatial_index=spatial_index,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)


@skeletongroup.command("merge")
@click.argument("path")
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--min-cable-length', default=1000, help="Skip objects smaller than this physical path length. Default: 1000 nm", type=float)
@click.option('--max-cable-length', default=None, help="Skip objects larger than this physical path length. Default: no limit", type=float)
@click.option('--tick-threshold', default=0, help="Remove small \"ticks\", or branches from the main skeleton one at a time from smallest to largest. Branches larger than this are preserved. Default: no elimination", type=float)
@click.option('--delete-fragments', is_flag=True, default=False, help="Delete the skeleton fragments from the first pass after upload is complete.")
@click.option('--magnitude', default=2, help="Split up the work with 10^(magnitude) prefix based tasks. Default: 2 (100 tasks)", type=int)
@click.pass_context
def skeleton_merge(
  ctx, path, queue, 
  min_cable_length, max_cable_length, 
  tick_threshold, delete_fragments, 
  magnitude
):
  """
  (2) Postprocess fragments into finished skeletons.
  """
  tasks = tc.create_unsharded_skeleton_merge_tasks(
    path, 
    magnitude=magnitude, 
    dust_threshold=min_cable_length,
    max_cable_length=max_cable_length,
    tick_threshold=tick_threshold,
    delete_fragments=delete_fragments,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@skeletongroup.command("merge-sharded")
@click.argument("path")
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--min-cable-length', default=1000, help="Skip objects smaller than this physical path length. Default: 1000 nm", type=float)
@click.option('--max-cable-length', default=None, help="Skip objects larger than this physical path length. Default: no limit", type=float)
@click.option('--tick-threshold', default=0, help="Remove small \"ticks\", or branches from the main skeleton one at a time from smallest to largest. Branches larger than this are preserved. Default: no elimination", type=float)
@click.option('--preshift-bits', default=3, help="Shift LSB to try to increase number of hash collisions.", type=int)
@click.option('--minishard-bits', default=10, help="2^bits number of bays holding variable numbers of labels per shard.", type=int)
@click.option('--shard-bits', default=2, help="2^bits number of shard files to generate.", type=int)
@click.option('--minishard-index-encoding', default="gzip", help="Minishard indices can be compressed. gzip or raw. Default: gzip")
@click.option('--data-encoding', default="gzip", help="Shard data can be compressed. gzip or raw. Default: gzip")
@click.pass_context
def skeleton_sharded_merge(
  ctx, path, queue, 
  min_cable_length, max_cable_length, 
  tick_threshold, 
  preshift_bits, minishard_bits, shard_bits,
  minishard_index_encoding, data_encoding
):
  """
  (2) Postprocess fragments into finished skeletons.

  Only use this command if you used the --sharded flag
  during the forging step. Some reasonable defaults
  are selected for a dataset with a few million labels,
  but for smaller or larger datasets they may not be
  appropriate.
  """
  tasks = tc.create_sharded_skeleton_merge_tasks(
    path, 
    dust_threshold=min_cable_length,
    max_cable_length=max_cable_length,
    tick_threshold=tick_threshold,
    preshift_bits=preshift_bits, 
    minishard_bits=minishard_bits, 
    shard_bits=shard_bits,
    minishard_index_encoding=minishard_index_encoding, 
    data_encoding=data_encoding,
  )

  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@main.group("rm")
def deletegroup():
  """
  Parallelize the deletion process for
  different kinds of data.
  """
  pass

@deletegroup.command("image")
@click.argument("path")
@click.option('--queue', required=True, help="AWS SQS queue or directory to be used for a task queue. e.g. sqs://my-queue or ./my-queue. See https://github.com/seung-lab/python-task-queue", type=str)
@click.option('--mip', default=0, help="Which mip level to start deleting from. Default: 0")
@click.option('--num-mips', default=5, help="The number of mip levels to delete at once. Default: 5")
@click.option('--shape', default=None, help="The size of each deletion task as a comma delimited list. Must be a multiple of the chunk size.", type=Tuple3())
@click.pass_context
def delete_images(
  ctx, path, queue, 
  mip, num_mips, shape
):
  """
  Delete the image layer of a dataset.
  """
  tasks = tc.create_deletion_tasks(
    path, mip, num_mips=num_mips, shape=shape
  )
  parallel = int(ctx.obj.get("parallel", 1))
  tq = TaskQueue(normalize_path(queue))
  tq.insert(tasks, parallel=parallel)

@main.group("design")
def designgroup():
  """
  Tools to aid the design of tasks or neuroglancer layers.
  """
  pass


@designgroup.command("ds-memory")
@click.argument("path")
@click.argument("memory_bytes", type=float)
@click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
@click.option('--factor', default="2,2,1", type=Tuple3(), help="Downsample factor to use.", show_default=True)
@click.option('--verbose', is_flag=True, help="Print some additional information.")
@click.option('--max-mips', default=5, help="Maximum downsamples to generate from this shape.", show_default=True)
def dsmemory(path, memory_bytes, mip, factor, verbose, max_mips):
  """
  Compute the task shape that maximizes the number of
  downsamples for a given amount of memory.
  """ 
  cv = CloudVolume(path, mip=mip)

  data_width = np.dtype(cv.dtype).itemsize
  cx, cy, cz = cv.chunk_size
  memory_bytes = int(memory_bytes)

  shape = downsample_scales.downsample_shape_from_memory_target(
    data_width, 
    cx, cy, cz, factor, 
    memory_bytes, max_mips
  )

  num_downsamples = int(math.log2(max(shape / cv.chunk_size)))

  mem_used = memory_used(data_width, shape, factor)

  shape = [ str(x) for x in shape ]
  shape = ",".join(shape)

  if verbose:
    print(f"Data Width: {data_width} B")
    print(f"Factor: {factor}")
    print(f"Chunk Size: {cx}, {cy}, {cz} voxels")
    print(f"Memory Limit: {format_bytes(memory_bytes, True)}")
    print("-----")
    print(f"Optimized Shape: {shape}")
    print(f"Downsamples: {num_downsamples}")
    print(f"Memory Used*: {format_bytes(mem_used, True)}")
    print(
      "\n"
      "* memory used is for retaining the image "
      "and all downsamples.\nAdditional costs may be incurred "
      "from processing."
    )
  else:
    print(shape)

@designgroup.command("ds-shape")
@click.argument("path")
@click.argument("shape", type=Tuple3())
@click.option('--mip', default=0, help="Select level of the image pyramid.", show_default=True)
@click.option('--factor', default="2,2,1", type=Tuple3(), help="Downsample factor to use.", show_default=True)
def dsshape(path, shape, mip, factor):
  """
  Compute the approximate memory usage for a
  given downsample task shape.
  """ 
  cv = CloudVolume(path, mip=mip)
  data_width = np.dtype(cv.dtype).itemsize
  memory_bytes = memory_used(data_width, shape, factor)  
  print(format_bytes(memory_bytes, True))

def memory_used(data_width, shape, factor): 
  memory_bytes = data_width * shape[0] * shape[1] * shape[2]

  if factor not in ((1,1,1), (2,2,1), (2,1,2), (1,2,2), (2,2,2)):
    print(f"igneous: factor must be 2,2,1 or 2,2,2. Got: {factor}")
    sys.exit()

  # comes from a converging infinite series proof
  constant = factor[0] * factor[1] * factor[2]
  if constant != 1:
    memory_bytes *= constant / (constant - 1)

  return memory_bytes



import os
os.environ['GOOGLE_CLOUD_DISABLE_GRPC'] = 'true'

import socket

from google.cloud.logging.client import Client
from google.cloud.logging.logger import Logger

from igneous.secrets import PROJECT_NAME, QUEUE_NAME, google_credentials_path

client = Client.from_service_account_json(
    google_credentials_path, project=PROJECT_NAME)
logger =  Logger('pipeline_logger', client)

def log(severity, task, message):
    # Look at the log produce when running this script at:
    # https://console.cloud.google.com/logs/viewer?project=neuromancer-seung-import&resource=global
    # Choosing the severity:
    # DEBUG     Debug or trace information.
    # INFO      Routine information, such as ongoing status or performance.
    # NOTICE    Normal but significant events, such as start up, shut down, or a configuration change.
    # WARNING   Warning events might cause problems.
    # ERROR     Error events are likely to cause problems.
    # CRITICAL  Critical events cause more severe problems or outages.
    # ALERT     A person must take an action immediately.
    # EMERGENCY One or more systems are unusable.
    
    #TODO change resource from global to GKE containter or similar
    if hasattr(task,'_id'):
        task_id = task._id
        task_str = task.__repr__()
    else:
        task_id = 'undefined'
        task_str = 'undefined task'

    extended_message = '{}\n {}\n on host {}'.format(message, task_str, socket.gethostname())
    print(extended_message)

    logger.log_struct({'message': extended_message, 
                       'taskName':task_id, 
                       'taskQueueName':QUEUE_NAME},
                       severity=severity)




from __future__ import print_function

import os

def toversion(s):
  s = s.lower()
  allowed = set('abcdefghijklmnopqrstuvwxyz' + '0123456789' + '-')
  s = [ x for x in s if x in allowed ]
  return ''.join(s)

print("export APPVERSION="+toversion(os.environ['TRAVIS_BRANCH']))


from functools import reduce, partial
from typing import Any, Dict, Tuple, cast

from time import strftime

import numpy as np

import cloudvolume
import cloudvolume.exceptions
from cloudvolume import CloudVolume
from cloudvolume.lib import Vec, Bbox, max2, min2, xyzrange, find_closest_divisor, yellow, jsonify
from cloudfiles import CloudFiles

from igneous.tasks import (
  MeshTask, MeshManifestTask, GrapheneMeshTask
)
from .common import (
  operator_contact, FinelyDividedTaskIterator, 
  get_bounds, num_tasks, graphene_prefixes
)

__all__ = [
  "create_meshing_tasks",
  "create_mesh_manifest_tasks",
  "create_graphene_meshing_tasks",
  "create_graphene_hybrid_mesh_manifest_tasks",
]

# split the work up into ~1000 tasks (magnitude 3)
def create_mesh_manifest_tasks(layer_path, magnitude=3, mesh_dir=None):
  assert int(magnitude) == magnitude

  start = 10 ** (magnitude - 1)
  end = 10 ** magnitude

  class MeshManifestTaskIterator(object):
    def __len__(self):
      return 10 ** magnitude
    def __iter__(self):
      for prefix in range(1, start):
        yield MeshManifestTask(layer_path=layer_path, prefix=str(prefix) + ':', mesh_dir=mesh_dir)

      # enumerate from e.g. 100 to 999
      for prefix in range(start, end):
        yield MeshManifestTask(layer_path=layer_path, prefix=prefix, mesh_dir=mesh_dir)

  return MeshManifestTaskIterator()


def create_meshing_tasks(
    layer_path, mip, shape=(448, 448, 448), 
    simplification=True, max_simplification_error=40,
    mesh_dir=None, cdn_cache=False, dust_threshold=None,
    object_ids=None, progress=False, fill_missing=False,
    encoding='precomputed', spatial_index=True, sharded=False,
    compress='gzip'
  ):
  shape = Vec(*shape)

  vol = CloudVolume(layer_path, mip)

  if mesh_dir is None:
    mesh_dir = 'mesh_mip_{}_err_{}'.format(mip, max_simplification_error)

  if not 'mesh' in vol.info:
    vol.info['mesh'] = mesh_dir
    vol.commit_info()

  cf = CloudFiles(layer_path)
  info_filename = '{}/info'.format(mesh_dir)
  mesh_info = cf.get_json(info_filename) or {}
  mesh_info['@type'] = 'neuroglancer_legacy_mesh'
  mesh_info['mip'] = int(vol.mip)
  mesh_info['chunk_size'] = shape.tolist()
  if spatial_index:
    mesh_info['spatial_index'] = {
        'resolution': vol.resolution.tolist(),
        'chunk_size': (shape*vol.resolution).tolist(),
    }
  cf.put_json(info_filename, mesh_info)

  class MeshTaskIterator(FinelyDividedTaskIterator):
    def task(self, shape, offset):
      return MeshTask(
        shape=shape.clone(),
        offset=offset.clone(),
        layer_path=layer_path,
        mip=vol.mip,
        simplification_factor=(0 if not simplification else 100),
        max_simplification_error=max_simplification_error,
        mesh_dir=mesh_dir, 
        cache_control=('' if cdn_cache else 'no-cache'),
        dust_threshold=dust_threshold,
        progress=progress,
        object_ids=object_ids,
        fill_missing=fill_missing,
        encoding=encoding,
        spatial_index=spatial_index,
        sharded=sharded,
        compress=compress,
      )

    def on_finish(self):
      vol.provenance.processing.append({
        'method': {
          'task': 'MeshTask',
          'layer_path': layer_path,
          'mip': vol.mip,
          'shape': shape.tolist(),
          'simplification': simplification,
          'max_simplification_error': max_simplification_error,
          'mesh_dir': mesh_dir,
          'fill_missing': fill_missing,
          'cdn_cache': cdn_cache,
          'dust_threshold': dust_threshold,
          'encoding': encoding,
          'object_ids': object_ids,
          'spatial_index': spatial_index,
          'sharded': sharded,
          'compress': compress,
        },
        'by': operator_contact(),
        'date': strftime('%Y-%m-%d %H:%M %Z'),
      }) 
      vol.commit_provenance()

  return MeshTaskIterator(vol.mip_bounds(mip), shape)

def create_graphene_meshing_tasks(
  cloudpath, timestamp, mip,
  simplification=True, max_simplification_error=40,
  mesh_dir=None, cdn_cache=False, object_ids=None, 
  progress=False, fill_missing=False, sharding=None,
  draco_compression_level=1, bounds=None
):
  cv = CloudVolume(cloudpath, mip=mip)

  if mip < cv.meta.watershed_mip:
    raise ValueError("Must mesh at or above the watershed mip level. Watershed MIP: {} Got: {}".format(
      cv.meta.watershed_mip, mip
    ))

  if mesh_dir is None:
    mesh_dir = 'meshes'

  cv.info['mesh'] = mesh_dir # necessary to set the mesh.commit_info() dir right
  if not 'mesh' in cv.info:
    cv.commit_info()

  watershed_downsample_ratio = cv.resolution // cv.meta.resolution(cv.meta.watershed_mip)
  shape = Vec(*cv.meta.graph_chunk_size) // watershed_downsample_ratio

  cv.mesh.meta.info['@type'] = 'neuroglancer_legacy_mesh'
  cv.mesh.meta.info['mip'] = cv.mip
  cv.mesh.meta.info['chunk_size'] = list(shape)
  if sharding:
    cv.mesh.meta.info['sharding'] = sharding
  cv.mesh.meta.commit_info()

  simplification = (0 if not simplification else 100)

  class GrapheneMeshTaskIterator(FinelyDividedTaskIterator):
    def task(self, shape, offset):
      return GrapheneMeshTask(
        cloudpath=cloudpath,
        shape=shape.clone(),
        offset=offset.clone(),
        mip=int(mip),
        simplification_factor=simplification,
        max_simplification_error=max_simplification_error,
        draco_compression_level=draco_compression_level,
        mesh_dir=mesh_dir, 
        cache_control=('' if cdn_cache else 'no-cache'),
        progress=progress,
        fill_missing=fill_missing,
        timestamp=timestamp,
      )

    def on_finish(self):
      cv.provenance.processing.append({
        'method': {
          'task': 'GrapheneMeshTask',
          'cloudpath': cv.cloudpath,
          'shape': cv.meta.graph_chunk_size,
          'mip': int(mip),
          'simplification': simplification,
          'max_simplification_error': max_simplification_error,
          'mesh_dir': mesh_dir,
          'fill_missing': fill_missing,
          'cdn_cache': cdn_cache,
          'timestamp': timestamp,
          'draco_compression_level': draco_compression_level,
        },
        'by': operator_contact(),
        'date': strftime('%Y-%m-%d %H:%M %Z'),
      }) 
      cv.commit_provenance()

  if bounds is None:
    bounds = cv.meta.bounds(mip).clone()
  else:
    bounds = cv.bbox_to_mip(bounds, mip=0, to_mip=mip)
    bounds = Bbox.clamp(bounds, cv.bounds)

  bounds = bounds.expand_to_chunk_size(shape, cv.voxel_offset)

  return GrapheneMeshTaskIterator(bounds, shape)

def create_graphene_hybrid_mesh_manifest_tasks(
  cloudpath, mip, mip_bits, x_bits, y_bits, z_bits
):
  prefixes = graphene_prefixes(mip, mip_bits, (x_bits, y_bits, z_bits))

  class GrapheneHybridMeshManifestTaskIterator(object):
    def __len__(self):
      return len(prefixes)
    def __iter__(self):
      for prefix in prefixes:
        yield MeshManifestTask(layer_path=cloudpath, prefix=str(prefix))

  return GrapheneHybridMeshManifestTaskIterator()



#!/usr/bin/python

"""
This script can be used to determine the actual bounding box
of a given layer. When a layer is first generated, it can
come from any pipeline. e.g. ChunkFlow.jl or another custom
method. These other methods are often agnostic to configuring
neuroglancer info files correctly, which is critical to viewing
the generated layer in neuroglancer and for processing them with
the pipeline. This script can be run to find the right parameters
for the info file.

Example Command:

  python compute_bounds.py gs://neuroglancer/DATASET/LAYER/

Example Output to stdout:

  bounds=Bbox([10,10,10], [138,138,138]) (size: [128,128,128]); chunk_size=[64,64,64]
  
"""

# this could be an igneous CLI command

import sys
import os

from tqdm import tqdm

from cloudvolume.lib import max2
from cloudvolume import CloudVolume, Storage, Bbox

from cloudfiles import CloudFiles

layer_path = sys.argv[1]

cv = CloudVolume(layer_path)
cf = CloudFiles(layer_path)

bboxes = []
for filename in tqdm(cf.list(prefix=cv.key), desc="Computing Bounds"):
  bboxes.append( Bbox.from_filename(filename) )

bounds = Bbox.expand(*bboxes)
chunk_size = reduce(max2, map(lambda bbox: bbox.size3(), bboxes))
print('bounds={} (size: {}); chunk_size={}'.format(bounds, bounds.size3(), chunk_size))










from __future__ import print_function

import os
import re
import subprocess
import sys

from cloudvolume import Storage, DataLayerProvenance

def ls(cloudpath):
  listing = subprocess.check_output(['gsutil', 'ls', cloudpath])
  listing = listing.split('\n')
  return [ x for x in listing if x ]

valid_paths = re.compile(r'^(gs|file|s3)://([/\d\w_\.\-]+)/([\d\w_\.\-]+)/([\d\w_\.\-]+)/?')

datasets = ls('gs://neuroglancer') + ls('s3://neuroglancer')

missing_report = []
invalid_report = []
success_report = []

for dataset in datasets:
  layers = ls(dataset)

  for layer in layers:
    if not valid_paths.match(layer):
      continue 

    if 'removeme' in layer:
      continue

    with Storage(layer, n_threads=0) as stor:
      if not stor.exists('provenance'):
        missing_report.append(layer)
      else:
        prov = stor.get_file('provenance')

        try:
          prov = DataLayerProvenance().from_json(prov)
        except:
          invalid_report.append(layer)
        else:
          success_report.append(layer)

RESET_COLOR = "\033[m"
YELLOW = "\033[1;93m"
RED = '\033[1;91m'        
GREEN = '\033[1;92m' 

def colorize(color, array):
  print(color + "\n".join(array) + RESET_COLOR + "\n")

print("""
The following reports the status of the 'provenance' file in
each layer of a neuroglancer dataset. 
""")

if len(success_report):
  print("VALID")
  colorize(GREEN, success_report)

if len(invalid_report):
  print("INVALID")
  colorize(YELLOW, invalid_report)

if len(missing_report):
  print("MISSING")
  colorize(RED, missing_report)

print('done.')






import copy
from functools import reduce, partial
import operator
import os
import subprocess

import numpy as np

from cloudvolume.lib import Vec, Bbox, max2, min2, xyzrange, find_closest_divisor, yellow, jsonify

def operator_contact():
  contact = ''
  try:
    contact = subprocess.check_output("git config user.email", shell=True)
    contact = str(contact.rstrip())
  except:
    try:
      print(yellow('Unable to determine provenance contact email. Set "git config user.email". Using unix $USER instead.'))
      contact = os.environ['USER']
    except:
      print(yellow('$USER was not set. The "owner" field of the provenance file will be blank.'))
      contact = ''

  return contact

def prod(x):
  return reduce(operator.mul, x, 1)

def get_bounds(vol, bounds, mip, chunk_size=None):
  if bounds is None:
    bounds = vol.bounds.clone()
  else:
    bounds = Bbox.create(bounds)
    bounds = vol.bbox_to_mip(bounds, mip=0, to_mip=mip)
    if chunk_size is not None:
      bounds = bounds.expand_to_chunk_size(chunk_size, vol.meta.voxel_offset(mip))
    bounds = Bbox.clamp(bounds, vol.meta.bounds(mip))
  

  print("Volume Bounds: ", vol.meta.bounds(mip))
  print("Selected ROI:  ", bounds)

  return bounds

def num_tasks(bounds, shape):
  return int(reduce(operator.mul, np.ceil(bounds.size3() / shape)))

class FinelyDividedTaskIterator():
  """
  Parallelizes tasks that do not have overlap.

  Evenly splits tasks between processes without 
  regards to whether the dividing line lands in
  the middle of a slice. 
  """
  def __init__(self, bounds, shape):
    self.bounds = bounds 
    self.shape = Vec(*shape)
    self.start = 0
    self.end = num_tasks(bounds, shape)

  def __len__(self):
    return self.end - self.start
  
  def __getitem__(self, slc):
    itr = copy.deepcopy(self)
    itr.start = max(self.start + slc.start, self.start)
    itr.end = min(self.start + slc.stop, self.end)
    return itr

  def __iter__(self):
    for i in range(self.start, self.end):
      pt = self.to_coord(i)
      offset = pt * self.shape + self.bounds.minpt
      yield self.task(self.shape.clone(), offset.clone())

    self.on_finish()

  def to_coord(self, index):
    """Convert an index into a grid coordinate defined by the task shape."""
    sx, sy, sz = np.ceil(self.bounds.size3() / self.shape).astype(int)
    sxy = sx * sy
    z = index // sxy
    y = (index - (z * sxy)) // sx
    x = index - sx * (y + z * sy)
    return Vec(x,y,z)

  def task(self, shape, offset):
    raise NotImplementedError()

  def on_finish(self):
    pass

def graphene_prefixes(
    mip=1, mip_bits=8, 
    coord_bits=(10, 10, 10), 
    prefix_length=6
  ):
  """
  Graphene structures segids as decimal numbers following
  the below format:

  mip x y z segid

  Typical parameter values are 
  mip_bits=4 or 8, x_bits=8 or 10, y_bits=8 or 10
  """
  coord_bits = Vec(*coord_bits)

  mip_shift = 64 - mip_bits
  x_shift = mip_shift - coord_bits.x
  y_shift = x_shift - coord_bits.y
  z_shift = y_shift - coord_bits.z

  x_range = 2 ** coord_bits.x 
  y_range = 2 ** coord_bits.y
  z_range = 2 ** coord_bits.z

  prefixes = set()
  for x in range(x_range):
    for y in range(y_range):
      num = (mip << mip_shift) + (x << x_shift) + (y << y_shift)
      num = str(num)[:prefix_length]
      prefixes.add(num)

  return prefixes


from .skeleton import SkeletonTask, UnshardedSkeletonMergeTask, ShardedSkeletonMergeTask
from .mesh import MeshTask, MeshManifestTask, GrapheneMeshTask
from .image import (
  HyperSquareConsensusTask, #HyperSquareTask,
  DownsampleTask, QuantizeTask, 
  TransferTask, WatershedRemapTask, DeleteTask, 
  LuminanceLevelsTask, ContrastNormalizationTask,
  MaskAffinitymapTask, InferenceTask, BlackoutTask,
  TouchTask, ImageShardTransferTask
)